package dev.fitko.fitconnect.client.subscriber;

import com.nimbusds.jose.jwk.JWK;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.callback.Callback;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionState;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.metadata.payment.PaymentInformation;
import dev.fitko.fitconnect.api.domain.model.reply.CreatedReply;
import dev.fitko.fitconnect.api.domain.model.reply.SentReply;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.sender.SendableReply;
import dev.fitko.fitconnect.api.domain.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.client.util.ReplyChannelKeyGenerator;
import dev.fitko.fitconnect.client.util.ValidDataGuard;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.net.URI;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class ReplySenderTest {

    private FitConnectService serviceMock;

    private ReplySender underTest;

    @BeforeEach
    void setUp() {
        final ApplicationConfig config = new ApplicationConfig();
        serviceMock = mock(FitConnectService.class);
        underTest = new ReplySender(config, serviceMock, new ValidDataGuard(serviceMock));
    }

    @Test
    void testSendValidReply() {

        // Given
        final ServiceType serviceType = new ServiceType();
        serviceType.setName("Test");
        serviceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final Submission submission = new Submission();
        submission.setCaseId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setSubmissionId(UUID.randomUUID());
        submission.setServiceType(serviceType);
        submission.setCallback(new Callback());
        submission.setEncryptedData("encryptedData");
        submission.setEncryptedMetadata("encryptedMetadata");

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final var destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setServices(Set.of(destinationService));

        final JWK encryptionKey = ReplyChannelKeyGenerator.generateKeyPair().getPublicKey();

        final Metadata metadata = new Metadata();
        metadata.setReplyChannel(ReplyChannel.fromFitConnect(encryptionKey, List.of("https://processing.standard.net")));
        metadata.setPaymentInformation(new PaymentInformation());
        metadata.setAuthenticationInformation(List.of());

        final CreatedReply createdReply = new CreatedReply();
        createdReply.setReplyId(UUID.randomUUID());
        createdReply.setCaseId(UUID.randomUUID());

        final SentReply expectedSentReply = new SentReply(createdReply.getReplyId(), createdReply.getCaseId(),List.of(UUID.randomUUID()));

        when(serviceMock.getDestination(any())).thenReturn(destination);
        when(serviceMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());
        when(serviceMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());
        when(serviceMock.announceReply(any())).thenReturn(createdReply);
        when(serviceMock.encryptBytes(encryptionKey.toRSAKey(), "{\"test\" : \"data\" }".getBytes(), MimeType.APPLICATION_JSON.value())).thenReturn("encryptedData");
        when(serviceMock.encryptBytes(encryptionKey.toRSAKey(),"test".getBytes(), "text/plain")).thenReturn("encryptedAttachment");
        when(serviceMock.encryptObject(any(), any(), any())).thenReturn("encryptedMetadata");
        when(serviceMock.createHash(any())).thenReturn("hashedData");
        when(serviceMock.submitReply(any(), any())).thenReturn(expectedSentReply);

        final var submitState = new SubmissionStatus(SubmissionState.SUBMITTED, new Date(), List.of());
        final ReceivedSubmission receivedSubmission = new ReceivedSubmission(null, submission, metadata, null, Collections.emptyList(), Collections.emptyMap(), submitState);

        final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                .setJsonData("{\"test\" : \"data\" }", URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .addAttachment(Attachment.fromByteArray("test".getBytes(), "text/plain"))
                .build();

        // When
        final SentReply sentReply = underTest.sendReply(sendableReply);

        // Then
        assertNotNull(sentReply);
        assertThat(sentReply.getReplyId(), is(expectedSentReply.getReplyId()));
        assertThat(sentReply.getCaseId(), is(expectedSentReply.getCaseId()));
        assertThat(sentReply.getAttachments().size(), is(1));
    }
}