package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.sender.SendableEncryptedSubmission;
import dev.fitko.fitconnect.api.domain.sender.SendableReply;
import dev.fitko.fitconnect.api.domain.sender.SendableSubmission;
import dev.fitko.fitconnect.api.domain.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import java.net.URI;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class ValidDataGuardTest {

    ValidDataGuard underTest;
    FitConnectService fitConnectService;

    @BeforeEach
    void setUp() {
        fitConnectService = mock(FitConnectService.class);
        underTest = new ValidDataGuard(fitConnectService);
    }

    @Nested
    class SendableSubmissionDataTests {

        @Test
        void testValidSubmissionPayload() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test")
                    .setJsonData("\"test\": \"json\"", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();

            // Then valid if no exception is thrown
            underTest.ensureValidDataPayload(sendableSubmission);
        }

        @Test
        void testMissingJsonData() {

            // Given
            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .setJsonData(null, URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Data is mandatory, but was null"));
        }

        @Test
        void testMissingXmlData() {

            // Given
            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .setXmlData(null, URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Data is mandatory, but was null"));
        }

        @Test
        void testMissingDestinationId() {

            // Given
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(null)
                    .setServiceType("name", "urn:de:fim:leika:leistung:99400048079000")
                    .setJsonData("{}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("DestinationId is mandatory, but was null"));
        }

        @Test
        void testInvalidServiceIdentifier() {

            // Given
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("name", "illegal:test:identifier")
                    .setJsonData("{\"test\" . \"data\"}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("LeikaKey has invalid format"));
        }

        @Test
        void testMissingServiceIdentifier() {

            // Given
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType(null, "name")
                    .setJsonData("{\"test\" . \"data\"}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Leika key is mandatory, but was null"));
        }

        @Test
        void testServiceIdentifierNotMatchingDestination() {

            // Given
            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:123456789101114", "Test")
                    .setJsonData("\"test\": \"json\"", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Provided service type 'urn:de:fim:leika:leistung:123456789101114' is not allowed by the destination"));
        }

        @Test
        void testMimeTypeNotMatchingDestination() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_XML);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test")
                    .setJsonData("\"test\": \"json\"", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Combination of provided MIME type 'application/json' and schema URI " +
                    "'https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json' is not allowed by the destination"));
        }

        @Test
        void testSchemaUriNotMatchingDestination() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000010_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test")
                    .setJsonData("\"test\": \"json\"", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Combination of provided MIME type 'application/json' and schema URI " +
                    "'https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json' is not allowed by the destination"));
        }
    }

    @Nested
    class SendableEncryptedSubmissionDataTests {

        @Test
        void testValidEncryptedSubmissionPayload() {

            // Given
            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final SendableEncryptedSubmission sendableEncryptedSubmission = SendableEncryptedSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                    .setEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .setEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .build();

            // Then valid if no exception is thrown
            underTest.ensureValidDataPayload(sendableEncryptedSubmission);
        }

        @Test
        void testMissingEncryptedData() {

            // Given
            final SendableEncryptedSubmission sendableEncryptedSubmission = SendableEncryptedSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .setEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .setEncryptedData(null)
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableEncryptedSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Encrypted data is mandatory, but was null"));
        }

        @Test
        void testMissingEncryptedMetadata() {

            // Given
            final SendableEncryptedSubmission sendableEncryptedSubmission = SendableEncryptedSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("Test", "urn:de:fim:leika:leistung:99400048079000")
                    .setEncryptedMetadata(null)
                    .setEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableEncryptedSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Encrypted metadata must not be null"));
        }

        @Test
        void testMissingDestinationId() {

            // Given
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final SendableEncryptedSubmission sendableEncryptedSubmission = SendableEncryptedSubmission.Builder()
                    .setDestination(null)
                    .setServiceType("name", "urn:de:fim:leika:leistung:99400048079000")
                    .setEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .setEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableEncryptedSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("DestinationId is mandatory, but was null"));
        }

        @Test
        void testInvalidServiceIdentifier() {

            // Given
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final SendableEncryptedSubmission sendableEncryptedSubmission = SendableEncryptedSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("name", "illegal:test:identifier")
                    .setEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .setEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableEncryptedSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("LeikaKey has invalid format"));
        }

        @Test
        void testMissingServiceIdentifier() {

            // Given
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final SendableEncryptedSubmission sendableEncryptedSubmission = SendableEncryptedSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType(null, "name")
                    .setEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .setEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .build();
            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableEncryptedSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Leika key is mandatory, but was null"));
        }

        @Test
        void testServiceIdentifierNotMatchingDestination() {

            // Given
            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

            final Destination destination = new Destination();
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("urn:something")))).thenReturn(ValidationResult.ok());

            final SendableEncryptedSubmission sendableEncryptedSubmission = SendableEncryptedSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:11111111111111", "name")
                    .setEncryptedMetadata("4Y0sJhadfrQnNZXeS7Pqh73FvtF4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .setEncryptedData("4Y0sJhadfrQnNZXeS7Pqh73FvtF")
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableEncryptedSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Provided service type 'urn:de:fim:leika:leistung:11111111111111' is not allowed by the destination"));
        }
    }

    @Nested
    class SendableReplyDataTests {

        @Test
        void testMissingFitConnectReplyChannel() {

            // Given
            final Metadata metadata = new Metadata();

            final Submission submission = new Submission();
            submission.setCaseId(UUID.randomUUID());
            submission.setSubmissionId(UUID.randomUUID());
            submission.setServiceType(new ServiceType("Test", "urn:de:fim:leika:leistung:99400048079000"));

            final ReceivedSubmission receivedSubmission = new ReceivedSubmission(fitConnectService, submission, metadata, null, null, null, null);

            final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                    .setJsonData("data", URI.create("https://test.url.net"))
                    .build();

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableReply));

            // Then
            assertThat(exception.getMessage(), containsString("FitConnect reply channel is mandatory for sending replies, but was null on original submission " + submission.getSubmissionId()));
        }

        @Test
        void testValidSendableReply() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            final Metadata metadata = new Metadata();
            metadata.setReplyChannel(ReplyChannel.fromFitConnect(ReplyChannelKeyGenerator.generateKeyPair().getPublicKey(), List.of("https://process.test-standard.net")));

            final Submission submission = new Submission();
            submission.setCaseId(UUID.randomUUID());
            submission.setSubmissionId(UUID.randomUUID());
            submission.setServiceType(new ServiceType("Test", "urn:de:fim:leika:leistung:99400048079000"));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final ReceivedSubmission receivedSubmission = new ReceivedSubmission(fitConnectService, submission, metadata, null, null, null, null);

            final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                    .setJsonData("{\"reply\" . \"tes-data\"}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();

            // valid if no exception occurs
            underTest.ensureValidDataPayload(sendableReply);
        }

        @Test
        void testReplyProcessingStandardsDoNotMatchDestination() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));
            service.setReplyChannels(ReplyChannel.fromFitConnect(ReplyChannelKeyGenerator.generateKeyPair().getPublicKey(), List.of("https://destination.test-standard.net")));

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            final Metadata metadata = new Metadata();
            metadata.setReplyChannel(ReplyChannel.fromFitConnect(ReplyChannelKeyGenerator.generateKeyPair().getPublicKey(), List.of("https://reply.test-standard.net")));

            final Submission submission = new Submission();
            submission.setCaseId(UUID.randomUUID());
            submission.setSubmissionId(UUID.randomUUID());
            submission.setServiceType(new ServiceType("Test", "urn:de:fim:leika:leistung:99400048079000"));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final ReceivedSubmission receivedSubmission = new ReceivedSubmission(fitConnectService, submission, metadata, null, null, null, null);

            final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                    .setJsonData("{\"reply\" : \"test data\"}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();

            // When / Then
            assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableReply));
        }

        @Test
        void testReplyProcessingStandardsDoMatchForMultipleDestinationServices() {

            // Given
            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service1 = new DestinationService();
            service1.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service1.setSubmissionSchemas(Set.of(submissionSchema));
            service1.setReplyChannels(ReplyChannel.fromFitConnect(ReplyChannelKeyGenerator.generateKeyPair().getPublicKey(), List.of("https://destination.test-standard-1.net")));

            final DestinationService service2 = new DestinationService();
            service2.setIdentifier("urn:de:fim:leika:leistung:1234567890");
            service2.setSubmissionSchemas(Set.of(submissionSchema));
            service2.setReplyChannels(ReplyChannel.fromFitConnect(ReplyChannelKeyGenerator.generateKeyPair().getPublicKey(), List.of("https://destination.test-standard-1.net")));

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service1, service2));

            final Metadata metadata = new Metadata();
            metadata.setReplyChannel(ReplyChannel.fromFitConnect(ReplyChannelKeyGenerator.generateKeyPair().getPublicKey(), List.of("https://destination.test-standard-1.net")));

            final Submission submission = new Submission();
            submission.setCaseId(UUID.randomUUID());
            submission.setSubmissionId(UUID.randomUUID());
            submission.setServiceType(new ServiceType("Test", "urn:de:fim:leika:leistung:99400048079000"));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final ReceivedSubmission receivedSubmission = new ReceivedSubmission(fitConnectService, submission, metadata, null, null, null, null);

            final SendableReply sendableReply = SendableReply.from(receivedSubmission)
                    .setJsonData("{\"reply\" : \"test data\"}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .build();

            // Then no exception is thrown
            underTest.ensureValidDataPayload(sendableReply);
        }

    }

    @Nested
    class MiscTests {

        @Test
        void testInvalidDestinationId() {

            // Given
            final var invalidUUID = "1234:132434:5678";

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () ->
                    SendableSubmission.Builder()
                            .setDestination(UUID.fromString(invalidUUID))
                            .setServiceType("urn:de:fim:leika:leistung:99400048079000", "name")
                            .setJsonData("{}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                            .build());

            // Then
            assertThat(exception.getMessage(), containsString("Invalid UUID string: " + invalidUUID));
        }

        @Test
        void testInvalidCallbackUri() {

            // Given
            final var invalidHttpScheme = URI.create("http://example.callback.net");

            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test")
                    .setJsonData("\"test\": \"json\"", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .setCallback(invalidHttpScheme, "longEnoughCallbackSecret")
                    .build();

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Callback URI " + invalidHttpScheme + " must be a secure https connection"));
        }

        @Test
        void testInvalidCallbackSecret() {

            // Given
            final var invalidCallbackSecret = "shortSecret";

            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));

            final DestinationService service = new DestinationService();
            service.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
            service.setSubmissionSchemas(Set.of(submissionSchema));

            final Destination destination = new Destination();
            destination.setDestinationId(UUID.randomUUID());
            destination.setServices(Set.of(service));

            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            final SendableSubmission sendableSubmission = SendableSubmission.Builder()
                    .setDestination(UUID.randomUUID())
                    .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test")
                    .setJsonData("\"test\": \"json\"", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                    .setCallback(URI.create("https://example.callback.net"), invalidCallbackSecret)
                    .build();


            when(fitConnectService.getDestination(any())).thenReturn(destination);
            when(fitConnectService.validateJsonFormat(any(), eq(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")))).thenReturn(ValidationResult.ok());

            // When
            final IllegalArgumentException exception = assertThrows(IllegalArgumentException.class, () -> underTest.ensureValidDataPayload(sendableSubmission));

            // Then
            assertThat(exception.getMessage(), containsString("Callback secret must have a length of at least 32 characters"));
        }


    }
}